/*****************************************************
 * Routines to get a move, validate a move,
 * apply a move to the board,
 * and detect the end of the game.
 *
 * Kenrick Mock 12/20/01
 *
 * 3/3/02 - Fixed bug in IsValid if tried to move
 *          piece to the same coordinates as source
 *****************************************************/
#include <stdio.h>
#include <string.h>
#include "globals.h"
#include "move.h"

/*****************************************************
 * IsValid
 *
 * Returns 1 if the move is a valid one, 0 otherwise.
 * whoseturn = 'b' or 'w' to indicate the current player.
 * x1,y1 = source coordinates for move
 * x2,y2 = destination coordinates for move
 *****************************************************/
int IsValid(char board[DIM][DIM], char whoseturn,
	    int x1, int y1, int x2, int y2) 
{
 int i=0, temp=0;

 /* Invalid if x1==x2 and y1==y2 */
 if ((x1==x2) && (y1==y2)) return 0;

 /* Invalid if any value is out of bounds */
 if ((x1<1) || (x1>=DIM) ||
     (y1<1) || (y1>=DIM) ||
     (x2<1) || (x2>=DIM) ||
     (y2<1) || (y2>=DIM)) return 0;

 /* Make sure there is a piece of the proper color at the source */
 if (whoseturn=='w') {
	if ((board[x1][y1]!=W_RUNNER) && (board[x1][y1]!=W_GUARD)) return 0;
 }
 else {
	if ((board[x1][y1]!=B_RUNNER) && (board[x1][y1]!=B_GUARD)) return 0;
 } 

 /* Check if move is valid */
 /* Move if this is a guard */
 if ((board[x1][y1]==W_GUARD) || (board[x1][y1]==B_GUARD)) {
	if (((x2 % 2)==1) || ((y2 % 2)==1)) return 0;   /* Dest must be even */
	/* Vertical move */
	if (x1==x2) {
		if (y1>y2) {	/* Moving up */
			for (i=y1-2; i>=y2; i-=2) {
				if (board[x1][i]!=EMPTY) return 0;
			}
			return 1;
		}
		else {		/* Moving down */
			for (i=y1+2; i<=y2; i+=2) {
				if (board[x1][i]!=EMPTY) return 0;
			}
			return 1;
		}	
	}
	/* Horizontal move */
	else if (y1==y2) {
		if (x1>x2) { 	/* Moving left */
			for (i=x1-2; i>=x2; i-=2) {
				if (board[i][y1]!=EMPTY) return 0;
			}
			return 1;
		}
		else {		/* Moving right */
			for (i=x1+2; i<=x2; i+=2) {
				if (board[i][y1]!=EMPTY) return 0;
			}
			return 1;
		}
	}
	else return 0;
 }

 /* Move if this is a white runner */
 else if (board[x1][y1]==W_RUNNER) {
	if (((x2 % 2)==0) || ((y2 % 2)==0)) return 0;   /* Dest must be odd */
	if (y1==y2) {		/* Side to side move */
		temp=x1-x2;
		if (temp<0) temp=temp*-1;
		if (temp>2) return 0;		/* Moved more than 1 square */
		if (board[x2][y2]==EMPTY) return 1;
	}
	/* Must be a upward move for white */
	if (x1!=x2) return 0;
	if (y1<y2) return 0;
	temp=y1-y2;
	if (temp<0) temp=temp*-1;
	if (temp>2) return 0;
	if (board[x2][y2]==EMPTY) return 1;
	else return 0;
 }

 /* Move if this is a black runner */
 else if (board[x1][y1]==B_RUNNER) {
	if (((x2 % 2)==0) || ((y2 % 2)==0)) return 0;   /* Dest must be odd */
	if (y1==y2) {		/* Side to side move */
		temp=x1-x2;
		if (temp<0) temp=temp*-1;
		if (temp>2) return 0;		/* Moved more than 1 square */
		if (board[x2][y2]==EMPTY) return 1;
	}
	/* Must be a downward move for white */
	if (x1!=x2) return 0;
	if (y1>y2) return 0;
	temp=y1-y2;
	if (temp<0) temp=temp*-1;
	if (temp>2) return 0;
	if (board[x2][y2]==EMPTY) return 1;
	else return 0;
 }

 return 1;
}

/***************************************************************************
 * RemoveRunner
 *
 * Returns 1 if the runner at the coordinates should be removed,
 *   0 otherwise.
 *   Assumes the input x and y are those for a valid runner.
 *   If verbose=1 we should output what happened and subtract
 *   from our global piece count
 **************************************************************************/
int RemoveRunner(char board[DIM][DIM], int x, int y, int verbose)
{
  int enemycount=0;
  char enemyGuard=W_GUARD;

  if (board[x][y]==W_RUNNER) enemyGuard=B_GUARD;
  if (x>1) {
	if (y>1) {
		if (board[x-1][y-1]==enemyGuard) enemycount++;
	}
	if (y<DIM-1) {
		if (board[x-1][y+1]==enemyGuard) enemycount++;
	}
  } 
  if (x<DIM-1) {
	if (y>1) {
		if (board[x+1][y-1]==enemyGuard) enemycount++;
	}
	if (y<DIM-1) {
		if (board[x+1][y+1]==enemyGuard) enemycount++;
	}
  } 

  if (enemycount>=2) {
	if (verbose) {
		printf("Runner at %d %d removed.\n",x,y);
  		if (board[x][y]==W_RUNNER) g_num_white--;
	 	else g_num_black--;	
	}
	return 1;
  }
  else return 0;
}


/*****************************************************
 * ApplyMove
 *
 * Takes a valid move and applies it to the board.
 * This moves the actual piece on the board and
 * removes any pieces that may be surrounded by two
 * enemy guards.
 *****************************************************/
void ApplyMove(char board[DIM][DIM], int x1, int y1, int x2, int y2, 
	int verbose)
{
 int i;
 char c, enemyRunner=B_RUNNER;

 /* Move the piece on the board */
 board[x2][y2]=board[x1][y1];
 board[x1][y1]=EMPTY;

 /* Remove any runners that are touching two or more enemy guards */
 c = board[x2][y2];
 if ((c==B_RUNNER) || (c==W_RUNNER)) {
	/* If a runner moved, it is the only piece that can be removed */
	if (RemoveRunner(board, x2, y2, verbose)) {
		board[x2][y2]=EMPTY;
	}
 } 
 else {
	/* If a guard moved, we need to check around the guard to
           see if any touching runners can be removed */
	if (c==B_GUARD) enemyRunner=W_RUNNER;
	if (board[x2-1][y2-1]==enemyRunner) {
		if (RemoveRunner(board, x2-1, y2-1, verbose)) {
			board[x2-1][y2-1]=EMPTY;
		}
	}
	if (board[x2-1][y2+1]==enemyRunner) {
		if (RemoveRunner(board, x2-1, y2+1, verbose)) {
			board[x2-1][y2+1]=EMPTY;
		}
	}
	if (board[x2+1][y2-1]==enemyRunner) {
		if (RemoveRunner(board, x2+1, y2-1, verbose)) {
			board[x2+1][y2-1]=EMPTY;
		}
	}
	if (board[x2+1][y2+1]==enemyRunner) {
		if (RemoveRunner(board, x2+1, y2+1, verbose)) {
			board[x2+1][y2+1]=EMPTY;
		}
	}
 }

}

/*****************************************************
 * GameOver
 *
 * Returns 1 if black won, 2 if white won, 0 otherwise.
 * Also prints out the winner if verbose is 1.  
 *****************************************************/
int GameOver(char board[DIM][DIM], int verbose)
{
 int w_runners=0, b_runners=0,i,j;
 
 for (i=1; i<DIM; i+=2) {
	if (board[i][1]==W_RUNNER) {
		if (verbose) printf("WHITE WINS!\n");
		return 2;
	}
	if (board[i][DIM-1]==B_RUNNER) {
		if (verbose) printf("BLACK WINS!\n");
		return 1;
	}
 }
 // Count up runners
 w_runners = g_num_white;
 b_runners = g_num_black;
/*
 for (j=1; j<DIM; j+=2) {
	for (i=1; i<DIM; i+=2) {
		if (board[i][j]==W_RUNNER) w_runners++;
		else if (board[i][j]==B_RUNNER) b_runners++;
	}
 }
*/

 if (w_runners==0) {
	if (verbose) printf("BLACK WINS! No white runners left.\n");
	return 1;
 }
 if (b_runners==0) {
	if (verbose) printf("WHITE WINS! No black runners left.\n");
	return 2;
 }
 return 0;
}



/*****************************************************
 * GetMove
 *
 * Gets a move from standard input.  The move must be
 * a valid one, or the user will be prompted to enter
 * the move again.
 *****************************************************/
void GetMove(char board[DIM][DIM], char whoseturn,
             int *x1, int *y1, int *x2, int *y2)
{
  if (whoseturn=='w') printf("\nWhite's move: ");
  else printf("\nBlack's move: " );
  do {
    printf("Enter src and dest coords (e.g., x1 y1 x2 y2): ");
    fflush(stdout);
    scanf("%d %d %d %d",x1,y1,x2,y2);
  } while (!IsValid(board, whoseturn, *x1, *y1, *x2, *y2));
}

/*****************************************************
 * GenerateMoves
 *
 *****************************************************/

void GenerateMoves(char board[DIM][DIM], char whoseturn, movelist *m)
{
 int i,j,k;
 char myGuard=W_GUARD;
 char myRunner=W_RUNNER;

 ClearMoveList(m);
 if (whoseturn=='b') { myGuard = B_GUARD; myRunner = B_RUNNER; }
 /* Check moving Runners */
 for (i=1; i<DIM; i+=2)
	for (j=1; j<DIM; j+=2) {
		if (board[i][j]==myRunner) {
			/* Move up */
			if ((whoseturn=='w') && (j>1) && 
		            (board[i][j-2]==EMPTY)) {
				// i,j,i,j-2 is valid
				AddMove(m,i,j,i,j-2);
			}
			/* Move down */
			if ((whoseturn=='b') && (j<DIM-1) && 
		            (board[i][j+2]==EMPTY)) {
				// i,j,i,j+2 is valid
				AddMove(m,i,j,i,j+2);
			}
			/* Move left */
			if ((i>1) && (board[i-2][j]==EMPTY)) {
				// i,j,i-2,j is valid
				AddMove(m,i,j,i-2,j);
			}
			/* Move right */
			if ((i<DIM-1) && (board[i+2][j]==EMPTY)) {
				// i,j,i+2,j is valid
				AddMove(m,i,j,i+2,j);
			}
		}
	}

 /* Check moving Guards */
 for (i=2; i<DIM; i+=2) 
	for (j=2; j<DIM; j+=2) {
		if (board[i][j]==myGuard) {
			/* Check left*/
		 	for (k=i-2; k>0; k-=2) {
				if (board[k][j]!=EMPTY) break;
				else {
					// i,j,k,j is valid
					AddMove(m,i,j,k,j);
				}
			}
			/* Check right */
		 	for (k=i+2; k<DIM-2; k+=2) {
				if (board[k][j]!=EMPTY) break;
				else {
					// i,j,k,j is valid
					AddMove(m,i,j,k,j);
				}
			}
			/* Check up */
		 	for (k=j-2; k>0; k-=2) {
				if (board[i][k]!=EMPTY) break;
				else {
					// i,j,i,k is valid
					AddMove(m,i,j,i,k);
				}
			}
			/* Check down */
		 	for (k=j+2; k<DIM-2; k+=2) {
				if (board[i][k]!=EMPTY) break;
				else {
					// i,j,i,k is valid
					AddMove(m,i,j,i,k);
				}
			}
		}
	}
}

